@@ -57,6 +57,7 @@ options:
         description:
           - List of image names to consider as cache source.
         type: list
+        elements: str
       dockerfile:
         description:
           - Use with state C(present) and source C(build) to provide an alternate name for the Dockerfile to use when building an image.
@@ -92,6 +93,11 @@ options:
           - Do not use cache when building an image.
         type: bool
         default: no
+      etc_hosts:
+        description:
+          - Extra hosts to add to C(/etc/hosts) in building containers, as a mapping of hostname to IP address.
+        type: dict
+        version_added: "2.9"
       args:
         description:
           - Provide a dictionary of C(key:value) build arguments that map to Dockerfile ARG directive.
@@ -126,6 +132,12 @@ options:
             be set in the container being built.
           - Needs Docker SDK for Python >= 3.7.0.
         type: bool
+      target:
+        description:
+          - When building an image specifies an intermediate build stage by
+            name as a final stage for the resulting image.
+        type: str
+        version_added: "2.9"
     version_added: "2.8"
   archive_path:
     description:
@@ -318,6 +330,7 @@ requirements:
 author:
   - Pavel Antonov (@softzilla)
   - Chris Houseknecht (@chouseknecht)
+  - Sorin Sbarnea (@ssbarnea)
 
 '''
 
@@ -413,6 +426,7 @@ image:
     sample: {}
 '''
 
+import errno
 import os
 import re
 import traceback
@@ -420,6 +434,7 @@ import traceback
 from distutils.version import LooseVersion
 
 from ansible.module_utils.docker.common import (
+    clean_dict_booleans_for_docker_api,
     docker_version,
     AnsibleDockerClient,
     DockerBaseClass,
@@ -465,9 +480,11 @@ class ImageManager(DockerBaseClass):
         self.load_path = parameters.get('load_path')
         self.name = parameters.get('name')
         self.network = build.get('network')
+        self.extra_hosts = clean_dict_booleans_for_docker_api(build.get('etc_hosts'))
         self.nocache = build.get('nocache', False)
         self.build_path = build.get('path')
         self.pull = build.get('pull')
+        self.target = build.get('target')
         self.repository = parameters.get('repository')
         self.rm = build.get('rm', True)
         self.state = parameters.get('state')
@@ -698,8 +715,8 @@ class ImageManager(DockerBaseClass):
                 if image and image['Id'] == self.results['image']['Id']:
                     self.results['changed'] = False
 
-                if push:
-                    self.push_image(repo, repo_tag)
+        if push:
+            self.push_image(repo, repo_tag)
 
     def build_image(self):
         '''
@@ -733,12 +750,16 @@ class ImageManager(DockerBaseClass):
             params['cache_from'] = self.cache_from
         if self.network:
             params['network_mode'] = self.network
+        if self.extra_hosts:
+            params['extra_hosts'] = self.extra_hosts
         if self.use_config_proxy:
             params['use_config_proxy'] = self.use_config_proxy
             # Due to a bug in docker-py, it will crash if
             # use_config_proxy is True and buildargs is None
             if 'buildargs' not in params:
                 params['buildargs'] = {}
+        if self.target:
+            params['target'] = self.target
 
         for line in self.client.build(**params):
             # line = json.loads(line)
@@ -765,22 +786,45 @@ class ImageManager(DockerBaseClass):
 
         :return: image dict
         '''
+        # Load image(s) from file
+        load_output = []
         try:
             self.log("Opening image %s" % self.load_path)
-            image_tar = open(self.load_path, 'rb')
+            with open(self.load_path, 'rb') as image_tar:
+                self.log("Loading image from %s" % self.load_path)
+                for line in self.client.load_image(image_tar):
+                    self.log(line, pretty_print=True)
+                    if "stream" in line or "status" in line:
+                        load_line = line.get("stream") or line.get("status") or ''
+                        load_output.append(load_line)
+        except EnvironmentError as exc:
+            if exc.errno == errno.ENOENT:
+                self.client.fail("Error opening image %s - %s" % (self.load_path, str(exc)))
+            self.client.fail("Error loading image %s - %s" % (self.name, str(exc)), stdout='\n'.join(load_output))
         except Exception as exc:
-            self.fail("Error opening image %s - %s" % (self.load_path, str(exc)))
+            self.client.fail("Error loading image %s - %s" % (self.name, str(exc)), stdout='\n'.join(load_output))
 
-        try:
-            self.log("Loading image from %s" % self.load_path)
-            self.client.load_image(image_tar)
-        except Exception as exc:
-            self.fail("Error loading image %s - %s" % (self.name, str(exc)))
+        # Collect loaded images
+        loaded_images = set()
+        for line in load_output:
+            if line.startswith('Loaded image:'):
+                loaded_images.add(line[len('Loaded image:'):].strip())
 
-        try:
-            image_tar.close()
-        except Exception as exc:
-            self.fail("Error closing image %s - %s" % (self.name, str(exc)))
+        if not loaded_images:
+            self.client.fail("Detected no loaded images. Archive potentially corrupt?", stdout='\n'.join(load_output))
+
+        expected_image = '%s:%s' % (self.name, self.tag)
+        if expected_image not in loaded_images:
+            self.client.fail(
+                "The archive did not contain image '%s'. Instead, found %s." % (
+                    expected_image, ', '.join(["'%s'" % image for image in sorted(loaded_images)])),
+                stdout='\n'.join(load_output))
+        loaded_images.remove(expected_image)
+
+        if loaded_images:
+            self.client.module.warn(
+                "The archive contained more images than specified: %s" % (
+                    ', '.join(["'%s'" % image for image in sorted(loaded_images)]), ))
 
         return self.client.find_image(self.name, self.tag)
 
@@ -788,7 +832,7 @@ class ImageManager(DockerBaseClass):
 def main():
     argument_spec = dict(
         source=dict(type='str', choices=['build', 'load', 'pull', 'local']),
-        build=dict(type='dict', suboptions=dict(
+        build=dict(type='dict', options=dict(
             cache_from=dict(type='list', elements='str'),
             container_limits=dict(type='dict', options=dict(
                 memory=dict(type='int'),
@@ -805,6 +849,8 @@ def main():
             rm=dict(type='bool', default=True),
             args=dict(type='dict'),
             use_config_proxy=dict(type='bool'),
+            target=dict(type='str'),
+            etc_hosts=dict(type='dict'),
         )),
         archive_path=dict(type='path'),
         container_limits=dict(type='dict', options=dict(
@@ -845,13 +891,21 @@ def main():
     def detect_build_network(client):
         return client.module.params['build'] and client.module.params['build'].get('network') is not None
 
+    def detect_build_target(client):
+        return client.module.params['build'] and client.module.params['build'].get('target') is not None
+
     def detect_use_config_proxy(client):
         return client.module.params['build'] and client.module.params['build'].get('use_config_proxy') is not None
 
+    def detect_etc_hosts(client):
+        return client.module.params['build'] and bool(client.module.params['build'].get('etc_hosts'))
+
     option_minimal_versions = dict()
     option_minimal_versions["build.cache_from"] = dict(docker_py_version='2.1.0', docker_api_version='1.25', detect_usage=detect_build_cache_from)
     option_minimal_versions["build.network"] = dict(docker_py_version='2.4.0', docker_api_version='1.25', detect_usage=detect_build_network)
+    option_minimal_versions["build.target"] = dict(docker_py_version='2.4.0', detect_usage=detect_build_target)
     option_minimal_versions["build.use_config_proxy"] = dict(docker_py_version='3.7.0', detect_usage=detect_use_config_proxy)
+    option_minimal_versions["build.etc_hosts"] = dict(docker_py_version='2.6.0', docker_api_version='1.27', detect_usage=detect_etc_hosts)
 
     client = AnsibleDockerClient(
         argument_spec=argument_spec,
